/*
const data =
{name: 'Root', children: [
  {name: 'A'},
  {name: 'B', children: [
    {name: 'B1'},
    {name: 'B2'},
    {name: 'B3'}
  ]},
  {name: 'C', children: [
    {name: 'C1'},
    {name: 'C2', children: [
      {name: 'C21'},
      {name: 'C22'},
      {name: 'C23'}
    ]}
  ]},
  {name: 'D'}
]}

const x = d3.hierarchy(data)
*/

declare module 'd3-hierarchy' {
  interface Node<T> {
    data:T;
    depth:number;
    height:number;
    parent:Node<T>;
    children:Node<T>[];
    value?:number;
    
    /**
     * ```
     * x.children[2].children[1].children[1].ancestors().map(z => z.data.name).join(' → ')
     * // C22 → C2 → C → Root
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_ancestors
     */
    ancestors():Node<T>[];
    
    /**
     * ```
     * x.descendants().map(z => z.data.name).join('\n')
     * // Root
     * // A
     * // B
     * // C
     * // D
     * // B1
     * // B2
     * // B3
     * // C1
     * // C2
     * // C21
     * // C22
     * // C23
     *
     * x.children[2].descendants().map(z => z.data.name).join('\n')
     * // C
     * // C1
     * // C2
     * // C21
     * // C22
     * // C23
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_descendants
     */
    descendants():Node<T>[];
    
    /**
     * ```
     * x.leaves().map(z => z.data.name).join('\n')
     * // A
     * // B1
     * // B2
     * // B3
     * // C1
     * // C21
     * // C22
     * // C23
     * // D
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_leaves
     */
    leaves():Node<T>[];
    
    /**
     * ```
     * x.children[2].children[1].children[1].path(x).map(z => z.data.name).join(' → ')
     * // C22 → C2 → C → Root
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_path
     */
    path(target:Node<T>):Node<T>[];
    
    /**
     * ```
     * x.links().map(z => `${z.source.data.name} → ${z.target.data.name}`).join('\n')
     * // Root → A
     * // Root → B
     * // Root → C
     * // Root → D
     * // B → B1
     * // B → B2
     * // B → B3
     * // C → C1
     * // C → C2
     * // C2 → C21
     * // C2 → C22
     * // C2 → C23
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_links
     */
    links():Link<T>[];
    
    /** @link https://github.com/d3/d3-hierarchy#node_sum */
    sum(fn:(d:T) => number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#node_sort */
    sort(compare:(a:T, b:T) => number):this;
    
    /**
     * ```
     * x.each(y => console.log(y.data.name))
     * // Root
     * // A
     * // B
     * // C
     * // D
     * // B1
     * // B2
     * // B3
     * // C1
     * // C2
     * // C21
     * // C22
     * // C23
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_each
     */
    each(fn:(d:Node<T>) => void):this;
    
    /**
     * ```
     * x.eachAfter(y => console.log(y.data.name))
     * // A
     * // B1
     * // B2
     * // B3
     * // B
     * // C1
     * // C21
     * // C22
     * // C23
     * // C2
     * // C
     * // D
     * // Root
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_eachAfter
     */
    eachAfter(fn:(d:Node<T>) => void):this;
    
    /**
     * ```
     * x.eachBefore(y => console.log(y.data.name))
     * // Root
     * // A
     * // B
     * // B1
     * // B2
     * // B3
     * // C
     * // C1
     * // C2
     * // C21
     * // C22
     * // C23
     * // D
     * ```
     * @link https://github.com/d3/d3-hierarchy#node_eachBefore
     */
    eachBefore(fn:(d:Node<T>) => void):this;
    
    /** @link https://github.com/d3/d3-hierarchy#node_copy */
    copy():Node<T>;
  }
  
  interface Link<T> {
    source:Node<T>;
    target:Node<T>;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#hierarchy */
  function hierarchy<T>(data:T, childrenSelector?:(d:T) => T[]):Node<T>;
  
  interface Stratify<T> {
    /** @link https://github.com/d3/d3-hierarchy#_stratify */
    (data:T[]):Node<T>;
    
    /** @link https://github.com/d3/d3-hierarchy#stratify_id */
    id():(d:T) => string;
    id(selector:(d:T) => string):this;
    
    /** @link https://github.com/d3/d3-hierarchy#stratify_parentId */
    parentId():(d:T) => string;
    parentId(selector:(d:T) => string):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#stratify */
  function stratify<T>():Stratify<T>;
  
  interface PointNode<T> extends Node<T> {
    x:number;
    y:number;
  }
  
  interface Cluster<T> {
    /** @link https://github.com/d3/d3-hierarchy#_cluster */
    (root:Node<T>):PointNode<T>;
    
    /** @link https://github.com/d3/d3-hierarchy#cluster_size */
    size():[number,number];
    size(size:[number,number]):this;
    
    /** @link https://github.com/d3/d3-hierarchy#cluster_nodeSize */
    nodeSize():[number,number];
    nodeSize(size:[number,number]):this;
    
    /** @link https://github.com/d3/d3-hierarchy#cluster_separation */
    separation():(a:Node<T>, b:Node<T>) => number;
    separation(separation:(a:Node<T>, b:Node<T>) => number):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#cluster */
  function cluster<T>():Cluster<T>;
  
  interface Tree<T> extends Cluster<T> {
  }
  
  /** @link https://github.com/d3/d3-hierarchy#tree */
  function tree<T>():Tree<T>;
  
  interface RectangleNode<T> extends Node<T> {
    x0:number;
    x1:number;
    y0:number;
    y1:number;
  }
  
  interface Treemap<T> {
    /** @link https://github.com/d3/d3-hierarchy#_treemap */
    (root:Node<T>):RectangleNode<T>;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_tile */
    tile():(node:Node<T>, x0:number, y0:number, x1:number, y1:number) => void;
    tile(tile:(node:Node<T>, x0:number, y0:number, x1:number, y1:number) => void):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_size */
    size():[number,number];
    size(size:[number,number]):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_round */
    round():boolean;
    round(round:boolean):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_padding */
    padding():(d:Node<T>) => number;
    padding(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingInner */
    paddingInner():(d:Node<T>) => number;
    paddingInner(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingOuter */
    paddingOuter():(d:Node<T>) => number;
    paddingOuter(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingTop */
    paddingTop():(d:Node<T>) => number;
    paddingTop(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingRight */
    paddingRight():(d:Node<T>) => number;
    paddingRight(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingBottom */
    paddingBottom():(d:Node<T>) => number;
    paddingBottom(padding:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#treemap_paddingLeft */
    paddingLeft():(d:Node<T>) => number;
    paddingLeft(padding:((d:Node<T>) => number) | number):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#treemap */
  function treemap<T>():Treemap<T>;
  
  /** @link https://github.com/d3/d3-hierarchy#treemapBinary */
  function treemapBinary(node:Node<any>, x0:number, y0:number, x1:number, y1:number);
  
  /** @link https://github.com/d3/d3-hierarchy#treemapDice */
  function treemapDice(node:Node<any>, x0:number, y0:number, x1:number, y1:number);
  
  /** @link https://github.com/d3/d3-hierarchy#treemapSlice */
  function treemapSlice(node:Node<any>, x0:number, y0:number, x1:number, y1:number);
  
  /** @link https://github.com/d3/d3-hierarchy#treemapSliceDice */
  function treemapSliceDice(node:Node<any>, x0:number, y0:number, x1:number, y1:number);
  
  interface Squarify {
    (node:Node<any>, x0:number, y0:number, x1:number, y1:number);
    
    /** @link https://github.com/d3/d3-hierarchy#squarify_ratio */
    ratio(ratio:number):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#treemapSquarify */
  const treemapSquarify:Squarify;
  
  /** @link https://github.com/d3/d3-hierarchy#treemapResquarify */
  const treemapResquarify:Squarify;
  
  interface Partition<T> {
    /** @link https://github.com/d3/d3-hierarchy#_partition */
    (root:Node<T>):RectangleNode<T>;
    
    /** @link https://github.com/d3/d3-hierarchy#partition_size */
    size():[number,number];
    size(size:[number,number]):this;
    
    /** @link https://github.com/d3/d3-hierarchy#partition_round */
    round():boolean;
    round(round:boolean):this;
    
    /** @link https://github.com/d3/d3-hierarchy#partition_padding */
    padding():(d:Node<T>) => number;
    padding(padding:((d:Node<T>) => number) | number):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#partition */
  function partition<T>():Partition<T>;
  
  interface CircleNode<T> extends Node<T> {
    x:number;
    y:number;
    r:number;
  }
  
  interface Pack<T> {
    /** @link https://github.com/d3/d3-hierarchy#_pack */
    (root:Node<T>):CircleNode<T>;
    
    /** @link https://github.com/d3/d3-hierarchy#pack_radius */
    radius():(d:Node<T>) => number;
    radius(radius:((d:Node<T>) => number) | number):this;
    
    /** @link https://github.com/d3/d3-hierarchy#pack_size */
    size():[number,number];
    size(size:[number,number]):this;
    
    /** @link https://github.com/d3/d3-hierarchy#pack_padding */
    padding():(d:Node<T>) => number;
    padding(padding:((d:Node<T>) => number) | number):this;
  }
  
  /** @link https://github.com/d3/d3-hierarchy#pack */
  function pack<T>():Pack<T>;
  
  /** @link https://github.com/d3/d3-hierarchy#packSiblings */
  function packSiblings<T extends {x:number,y:number,r:number}>(circles:{r:number}[]):T[];
  
  /** @link https://github.com/d3/d3-hierarchy#packEnclose */
  function packEnclose<T extends {x:number,y:number,r:number}>(circles:T[]):T;
}